# Author: Denis A. Engemann <denis.engemann@gmail.com>
# License: BSD (3-clause)

from . import viz
from . import preprocessing
from . import anatomy
from .anatomy import make_mne_anatomy, compute_forward_stack
from . import tests

from .io import read_raw
from .io import read_epochs
from .io import read_evokeds
from .io import read_info
from .io import read_annot
from .io import read_ica
from .io import read_trial_info

from .io import file_mapping

__version__ = '0.1.dev12'



# Author: Denis A. Engemann <denis.engemann@gmail.com>
# License: BSD (3-clause)

from . import config

import os
import os.path as op
from subprocess import call


def _download_testing_data():
    """download testing data
    .. note::
        requires python 2.7
    """
    for s3key in config.s3_keys:
        new_path = op.dirname(s3key).split(config.hcp_prefix)[-1][1:]
        new_path = op.join(config.hcp_path, new_path)
        fname = op.basename(s3key)
        new_file = op.join(new_path, fname)
        if not op.exists(new_path):
            os.makedirs(new_path)
        if not op.exists(new_file):
            print('downloading:\n\tfrom %s\n\tto %s' % (s3key, new_path))
            call(['s3cmd', 'get', s3key, new_path], shell=False)
            assert op.exists(new_file)



import hcp
from hcp.tests import config as tconf


def test_set_eog_channels():
    pass


def test_apply_ica():
    pass


def test_apply_ref_correction():
    pass


def test_map_ch_coords_to_mne():
    pass


def test_interpolate_missing():
    pass



from .file_mapping import get_file_paths, run_map


def get_s3_keys_anatomy(
        subject,
        freesurfer_outputs=('label', 'mri', 'surf'),
        meg_anatomy_outputs=('head_model', 'transforms'),
        mode='minimal',
        hcp_path_bucket='HCP_900'):
    """Helper to prepare AWS downloads for anatomy data

    A helper function useful for working with Amazon EC2 and S3.
    It compiles a list of related files.

    .. note::
        This function does not download anything.
        It only facilitates downloading by compiling the content.

    Parameters
    ----------
    subject : str
        The subject, a 6 digit code.
    freesurfer_outputs : str | list | tuple
        The Freesurfer outputs to be downloaded. Defaults to
        `('label', 'mri', 'surf')`.
    meg_anatomy_outputs : str | list | tuple
        The MEG anatomy contents to download. Defaults to
        `('head_model', 'transforms')`.
    mode : {'minimal', 'full'}
        Either all files or the most useful ones. Defaults to 'minimal'.
    hcp_path_bucket : str
        The S3 bucket path. Will be prepended to each file path.

    Returns
    -------
    aws_keys : list of str
        The AWS S3 keys to be downloaded.
    """
    aws_keys = list()
    for output in freesurfer_outputs:
        aws_keys.extend(
            get_file_paths(subject=subject, data_type='freesurfer',
                           output=output,
                           mode=mode,
                           hcp_path=hcp_path_bucket))
    for output in meg_anatomy_outputs:
        aws_keys.extend(
            get_file_paths(subject=subject, data_type='meg_anatomy',
                           output=output,
                           hcp_path=hcp_path_bucket))
    return aws_keys


def get_s3_keys_meg(
        subject, data_types, outputs=('raw', 'bads', 'ica'),
        run_inds=0, hcp_path_bucket='HCP_900', onsets='stim'):
    """Helper to prepare AWS downloads for MEG data

    A helper function useful for working with Amazon EC2 and S3.
    It compiles a list of related files.

    .. note::
        This function does not download anything.
        It only facilitates downloading by compiling the content.

    Parameters
    ----------
    subject : str
        The subject, a 6 digit code.
    data_type : str | tuple of str | list of str
        The acquisition context of the data. The following ones are supported:
        'rest'
        'noise'
        'task_motor'
        'task_story_math'
        'task_working_memory'
    outputs : str | tuple of str | list of str
        The kind of output. The following ones are supported:
        'raw'
        'epochs'
        'evoked'
        'ica'
        'annotations'
        'trial_info'
        'freesurfer'
        'meg_anatomy'
    onsets : {'stim', 'resp', 'sentence', 'block'} | list | tuple
        The event onsets. Only considered for epochs and evoked outputs
        The mapping is generous, everything that is not a response is a
        stimulus, in the sense of internal or external events. sentence and
        block are specific to task_story_math. Can be a collection of those.
    run_inds : int | list of int | tuple of int
        The run index. For the first run, use 0, for the second, use 1.
        Also see HCP documentation for the number of runs for a given data
        type.
    hcp_path_bucket : str
        The S3 bucket path. Will be prepended to each file path.

    Returns
    -------
    aws_keys : list of str
        The AWS S3 keys to be downloaded.
    """
    aws_keys = list()
    fun = get_file_paths
    if not isinstance(onsets, (list, tuple)):
        onsets = [onsets]
    if not isinstance(outputs, (list, tuple)):
        outputs = [outputs]
    if not isinstance(run_inds, (list, tuple)):
        run_inds = [run_inds]
    if not isinstance(data_types, (list, tuple)):
        data_types = [data_types]

    if not all(isinstance(rr, int) for rr in run_inds):
        raise ValueError('Rund indices must be integers. I found: ' +
                         ', '.join(['%s' % type(rr) for rr in run_inds
                                    if not isinstance(rr, int)]))
    elif max(run_inds) > 2:
        raise ValueError('For HCP MEG data there is no task with more'
                         'than three runs. Among your requests there '
                         'is run index %i. Did you forget about '
                         'zero-based indexing?' % max(run_inds))
    elif min(run_inds) < 0:
        raise ValueError('Run indices must be positive')

    for data_type in data_types:
        for output in outputs:
            if 'noise' in data_type and output != 'raw':
                continue  # we only have raw for noise data
            elif data_type == 'rest' and output in ('evoked', 'trial_info'):
                continue  # there is no such thing as evoked resting state data
            for run_index in run_inds:
                if run_index + 1 > len(run_map[data_type]):
                    continue  # ignore irrelevant run indices
                for onset in onsets:
                    aws_keys.extend(
                        fun(subject=subject, data_type=data_type,
                            output=output, run_index=run_index, onset=onset,
                            hcp_path=hcp_path_bucket))

    return aws_keys



import os
import os.path as op
import re

import numpy as np
from scipy import linalg

import mne
from mne.io.pick import _pick_data_channels, pick_info
from mne import write_trans, write_surface
from mne.transforms import Transform, apply_trans
from mne.utils import logger

from .io.file_mapping import get_file_paths
from .io.read import _read_trans_hcp
from .io.read import _get_head_model
from .io import read_info


def make_mne_anatomy(subject, subjects_dir, recordings_path=None,
                     hcp_path=op.curdir, mode='minimal', outputs=(
                         'label', 'mri', 'surf')):
    """Extract relevant anatomy and create MNE friendly directory layout

    The function will create the following outputs by default:

    $subjects_dir/$subject/bem/inner_skull.surf
    $subjects_dir/$subject/label/*
    $subjects_dir/$subject/mri/*
    $subjects_dir/$subject/surf/*
    $recordings_path/$subject/$subject-head_mri-trans.fif

    These can then be set as $SUBJECTS_DIR and as MEG directory, consistent
    with MNE examples.

    Parameters
    ----------
    subject : str
        The subject name.
    subjects_dir : str
        The path corresponding to MNE/freesurfer SUBJECTS_DIR (to be created)
    hcp_path : str
        The path where the HCP files can be found.
    mode : {'minimal', 'full'}
        If 'minimal', only the directory structure is created. If 'full' the
        freesurfer outputs shipped with HCP (see `outputs`) are symbolically
        linked.
    outputs : {'label', 'mri', 'stats', 'surf', 'touch'}
        The outputs of the freesrufer pipeline shipped by HCP. Defaults to
        ('mri', 'surf'), the minimum needed to extract MNE-friendly anatomy
        files and data.
    """
    if mode not in ('full', 'minimal'):
        raise ValueError('`mode` must either be "minimal" or "full"')
    if hcp_path == op.curdir:
        hcp_path = op.realpath(hcp_path)
    if not op.isabs(subjects_dir):
        subjects_dir = op.realpath(subjects_dir)

    this_subjects_dir = op.join(subjects_dir, subject)
    if not op.isabs(recordings_path):
        recordings_path = op.realpath(recordings_path)

    this_recordings_path = op.join(recordings_path, subject)

    if not op.exists(this_recordings_path):
        os.makedirs(this_recordings_path)

    for output in outputs:
        if not op.exists(op.join(this_subjects_dir, output)):
            os.makedirs(op.join(this_subjects_dir, output))
        if output == 'mri':
            for suboutput in ['orig', 'transforms']:
                if not op.exists(
                        op.join(this_subjects_dir, output, suboutput)):
                    os.makedirs(op.join(this_subjects_dir, output, suboutput))

        files = get_file_paths(
            subject=subject, data_type='freesurfer', output=output,
            mode=mode, hcp_path=hcp_path)
        for source in files:
            match = [match for match in re.finditer(subject, source)][-1]
            split_path = source[:match.span()[1] + 1]
            target = op.join(this_subjects_dir, source.split(split_path)[-1])
            if (not op.isfile(target) and not op.islink(target) and
                    op.exists(source)):  # don't link if it's not there.
                os.symlink(source, target)

    logger.info('reading extended structural processing ...')

    # Step 1 #################################################################
    # transform head models to expected coordinate system

    # make hcp trans
    transforms_fname = get_file_paths(
        subject=subject, data_type='meg_anatomy', output='transforms',
        hcp_path=hcp_path)
    transforms_fname = [k for k in transforms_fname if
                        k.endswith('transform.txt')][0]
    hcp_trans = _read_trans_hcp(fname=transforms_fname, convert_to_meter=False)

    # get RAS freesurfer trans
    c_ras_trans_fname = get_file_paths(
        subject=subject, data_type='freesurfer', output='mri',
        hcp_path=hcp_path)
    c_ras_trans_fname = [k for k in c_ras_trans_fname if
                         k.endswith('c_ras.mat')][0]
    logger.info('reading RAS freesurfer transform')
    # ceci n'est pas un .mat file ...

    with open(op.join(subjects_dir, c_ras_trans_fname)) as fid:
        ras_trans = np.array([
            r.split() for r in fid.read().split('\n') if r],
            dtype=np.float64)

    logger.info('Combining RAS transform and coregistration')
    ras_trans_m = linalg.inv(ras_trans)  # and the inversion

    logger.info('extracting head model')
    head_model_fname = get_file_paths(
        subject=subject, data_type='meg_anatomy', output='head_model',
        hcp_path=hcp_path)[0]
    pnts, faces = _get_head_model(head_model_fname=head_model_fname)

    logger.info('coregistring head model to MNE-HCP coordinates')
    pnts = apply_trans(ras_trans_m.dot(hcp_trans['bti2spm']), pnts)

    tri_fname = op.join(this_subjects_dir, 'bem', 'inner_skull.surf')
    if not op.exists(op.dirname(tri_fname)):
        os.makedirs(op.dirname(tri_fname))
    write_surface(tri_fname, pnts, faces)

    # Step 2 #################################################################
    # write corresponding device to MRI transform

    logger.info('extracting coregistration')
    # now convert to everything meter too here
    ras_trans_m[:3, 3] *= 1e-3
    bti2spm = hcp_trans['bti2spm']
    bti2spm[:3, 3] *= 1e-3
    head_mri_t = Transform(  # we're lying here for a good purpose
        'head', 'mri', np.dot(ras_trans_m, bti2spm))  # it should be 'ctf_head'
    write_trans(
        op.join(this_recordings_path, '%s-head_mri-trans.fif') % subject,
        head_mri_t)


def compute_forward_stack(subjects_dir,
                          subject,
                          recordings_path,
                          info_from=(('data_type', 'rest'), ('run_index', 0)),
                          fwd_params=None, src_params=None,
                          hcp_path=op.curdir, n_jobs=1):
    """
    Convenience function for conducting standard MNE analyses.

    .. note::
       this function computes bem solutions, source spaces and forward models
       optimized for connectivity computation, i.e., the fsaverage space
       is morphed onto the subject's space.

    Parameters
    ----------
    subject : str
        The subject name.
    hcp_path : str
        The directory containing the HCP data.
    recordings_path : str
        The path where MEG data and transformations are stored.
    subjects_dir : str
        The directory containing the extracted HCP subject data.
    info_from : tuple of tuples | dict
        The reader info concerning the data from which sensor positions
        should be read.
        Must not be empty room as sensor positions are in head
        coordinates for 4D systems, hence not available in that case.
        Note that differences between the sensor positions across runs
        are smaller than 12 digits, hence negligible.
    fwd_params : None | dict
        The forward parameters
    src_params : None | dict
        The src params. Defaults to:

        dict(subject='fsaverage', fname=None, spacing='oct6', n_jobs=2,
             surface='white', subjects_dir=subjects_dir, add_dist=True)
    hcp_path : str
        The prefix of the path of the HCP data.
    n_jobs : int
        The number of jobs to use in parallel.

    Returns
    -------
    out : dict
        A dictionary with the following keys:
            fwd : instance of mne.Forward
                The forward solution.
            src_subject : instance of mne.SourceSpace
                The source model on the subject's surface
            src_fsaverage : instance of mne.SourceSpace
                The source model on fsaverage's surface
            bem_sol : dict
                The BEM.
            info : instance of mne.io.meas_info.Info
                The actual measurement info used.
    """
    if isinstance(info_from, tuple):
        info_from = dict(info_from)

    head_mri_t = mne.read_trans(
        op.join(recordings_path, subject, '{}-head_mri-trans.fif'.format(
            subject)))

    src_params = _update_dict_defaults(
        src_params,
        dict(subject='fsaverage', fname=None, spacing='oct6', n_jobs=n_jobs,
             surface='white', subjects_dir=subjects_dir, add_dist=True))

    add_source_space_distances = False
    if src_params['add_dist']:  # we want the distances on the morphed space
        src_params['add_dist'] = False
        add_source_space_distances = True

    src_fsaverage = mne.setup_source_space(**src_params)
    src_subject = mne.morph_source_spaces(
        src_fsaverage, subject, subjects_dir=subjects_dir)

    if add_source_space_distances:  # and here we compute them post hoc.
        src_subject = mne.add_source_space_distances(
            src_subject, n_jobs=n_jobs)

    bems = mne.make_bem_model(subject, conductivity=(0.3,),
                              subjects_dir=subjects_dir,
                              ico=None)  # ico = None for morphed SP.
    bem_sol = mne.make_bem_solution(bems)
    bem_sol['surfs'][0]['coord_frame'] = 5

    info = read_info(subject=subject, hcp_path=hcp_path, **info_from)
    picks = _pick_data_channels(info, with_ref_meg=False)
    info = pick_info(info, picks)

    # here we assume that as a result of our MNE-HCP processing
    # all other transforms in info are identity
    for trans in ['dev_head_t', 'ctf_head_t']:
        #  'dev_ctf_t' is not identity
        assert np.sum(info[trans]['trans'] - np.eye(4)) == 0

    fwd = mne.make_forward_solution(
        info, trans=head_mri_t, bem=bem_sol, src=src_subject,
        n_jobs=n_jobs)

    return dict(fwd=fwd, src_subject=src_subject,
                src_fsaverage=src_fsaverage,
                bem_sol=bem_sol, info=info)


def _update_dict_defaults(values, defaults):
    """Helper to handle dict updates"""
    out = {k: v for k, v in defaults.items()}
    if isinstance(values, dict):
        out.update(values)
    return out



import os.path as op

"""Notes

For now:
- source pipeline are not considered
- datacheck pipelines are not considered
- EPRIME files are ignored

the following string formatters are used:

subject : the subject ID
run : the run ID
kind : the type of recording, e.g. 'Restin', 'Wrkmem'
pipeline : the name of the pipeline, e.g., 'icaclass'
context : either 'rmeg' or 'tmeg'
condition : the condition label for average pipelines
diff_modes : the contrast label for average pipelines, e.g. '[BT-diff]'
    or '[OP-diff]', etc.
sensor_mode : 'MODE-mag', 'MODE-grad'
"""

unprocessed = {
    'path': '{subject}/unprocessed/MEG/{run}-{kind}/4D',
    'patterns': ['c,rfDC', 'config'],
}

preprocessed = {
    'meg': {
        'path': '{subject}/MEG/{kind}/{pipeline}/',
        'patterns': {
            ('epochs', 'rmeg'): [
                '{subject}_MEG_{run}-{kind}_{context}preproc.mat'],
            ('epochs', 'tmeg'): [
                '{subject}_MEG_{run}-{kind}_tmegpreproc_{onset}.mat',
            ],
            'bads': [
                '{subject}_MEG_{run}-{kind}_baddata_badchannels.txt',
                '{subject}_MEG_{run}-{kind}_baddata_badsegments.txt',
                '{subject}_MEG_{run}-{kind}_baddata_manual_badchannels.txt',
                '{subject}_MEG_{run}-{kind}_baddata_manual_badsegments.txt'
            ],
            'ica': [
                '{subject}_MEG_{run}-{kind}_icaclass_vs.mat',
                '{subject}_MEG_{run}-{kind}_icaclass_vs.txt',
                '{subject}_MEG_{run}-{kind}_icaclass.mat',
                '{subject}_MEG_{run}-{kind}_icaclass.txt'
            ],
            'psd': ['{subject}_MEG_{run}-{kind}_powavg.mat'],
            'evoked': [
                (r'{subject}_MEG_{kind}_eravg_[{condition}]_{diff_modes}_'
                 '[{sensor_mode}].mat')
                ],
            'tfr': [
                (r'{subject}_MEG_{kind}_tfavg_[{condition}]_{diff_modes}_'
                 '[{sensor_mode}].mat')
            ],
            'trial_info': [
                '{subject}_MEG_{run}-{kind}_tmegpreproc_trialinfo.mat'
            ]
        }
    },
    'meg_anatomy': {
        'path': '{subject}/MEG/anatomy',
        'patterns': {
            'transforms': [
                '{subject}_MEG_anatomy_transform.txt',
            ],
            'head_model': [
                '{subject}_MEG_anatomy_headmodel.mat'
            ],
            'source_model': [
                '{subject}_MEG_anatomy_sourcemodel_2d.mat',
                '{subject}_MEG_anatomy_sourcemodel_3d4mm.mat',
                '{subject}_MEG_anatomy_sourcemodel_3d6mm.mat',
                '{subject}_MEG_anatomy_sourcemodel_3d8mm.mat'
            ],
            'freesurfer': [
                '{subject}.L.inflated.4k_fs_LR.surf.gii',
                '{subject}.R.inflated.4k_fs_LR.surf.gii',
                '{subject}.L.midthickness.4k_fs_LR.surf.gii']
        }
    },
    'freesurfer': {
        'path': '{subject}/T1w/{subject}',
        'patterns': {
            'label': [],
            'surf': [],
            'mri': [],
            'stats': [],
            'touch': []
        }
    }
}

file_modes = {
    'freesurfer':
        {'minimal': [
            'c_ras.mat',
            '001.mgz',
            'T1.mgz',
            'lh.white',
            'rh.white',
            'lh.inflated',
            'rh.inflated',
            'lh.sphere.reg',
            'rh.sphere.reg',
            'lh.sphere',
            'rh.sphere',
            'lh.curve',
            'rh.curve',
        ]},
}

evoked_map = {
    'modes': {'planar': 'MODE-planar', 'mag': 'MODE-mag'},
    # always BT-diff
    'task_motor': (
        'LM-TEMG-LF',
        'LM-TEMG-LH',
        'LM-TEMG-RF',
        'LM-TEMG-RH',
        'LM-TFLA-LF',
        'LM-TFLA-LH',
        'LM-TFLA-RF',
        'LM-TFLA-RH'),
    # if versus in name then OP-diff + BT-diff, else BT-diff
    'task_working_memory': (
        'LM-TIM-0B',
        'LM-TIM-0B-versus-2B',
        'LM-TIM-2B',
        'LM-TIM-face',
        'LM-TIM-face-versus-tool',
        'LM-TIM-tool',
        'LM-TRESP-0B',
        'LM-TRESP-0B-versus-2B',
        'LM-TRESP-2B',
        'LM-TRESP-face',
        'LM-TRESP-face-versus-tool',
        'LM-TRESP-tool'),
    # if versus in name then OP-diff + BT-diff, else BT-diff
    'task_story_math': (
        'LM-TEV-mathnumopt',
        'LM-TEV-mathnumoptcor-versus-mathnumoptwro',
        'LM-TEV-mathnumque',
        'LM-TEV-mathnumque-versus-mathoper',
        'LM-TEV-mathnumquelate-versus-mathnumqueearly',
        'LM-TEV-mathoper',
        'LM-TEV-mathsentnon',
        'LM-TEV-storoptcor-versus-storoptwro',
        'LM-TEV-storsentnon',
        'LM-TEV-storsentnon-versus-mathsentnon',
        'LM-TRESP-all')
}


freesurfer_files = op.join(op.dirname(__file__), 'data', '%s.txt')
for kind, patterns in preprocessed['freesurfer']['patterns'].items():
    with open(freesurfer_files % kind) as fid:
        patterns.extend([k.rstrip('\n') for k in fid.readlines()])

pipeline_map = {
    'ica': 'icaclass',
    'bads': 'baddata',
    'psd': 'powavg',
    'evoked': 'eravg',
    'tfr': 'tfavg',
    'trial_info': 'tmegpreproc'
}

kind_map = {
    'task_motor': 'Motort',
    'task_working_memory': 'Wrkmem',
    'task_story_math': 'StoryM',
    'rest': 'Restin',
    'noise_empty_room': 'Rnoise',
    'noise_subject': 'Pnoise',
    'meg_anatomy': 'anatomy',
    'freesurfer': 'freesurfer'
}

run_map = {
    'noise_empty_room': ['1'],
    'noise_subject': ['2'],
    'rest': ['3', '4', '5'],
    'task_working_memory': ['6', '7'],
    'task_story_math': ['8', '9'],
    'task_motor': ['10', '11'],
    'meg_anatomy': [],
    'freesurfer': []
}


def _map_onset(onset, data_type, output):
    """Helper to resolve stim and resp according to context"""
    out = onset
    if data_type == 'task_working_memory':
        out = {'stim': 'TIM', 'resp': 'TRESP'}[onset]
    elif data_type == 'task_motor':
        out = {'stim': 'TFLA', 'resp': 'TEMG'}[onset]
    elif data_type == 'task_story_math' and output == 'evoked':
        out = {'stim': 'TEV', 'resp': 'TRESP'}[onset]
    elif data_type == 'task_story_math' and output == 'epochs':
        out = {'stim': 'TEV', 'resp': 'TRESP', 'sentence': 'BSENT',
               'block': 'BUN'}[onset]
    return out


def _map_diff_mode(condition, data_type):
    """Helper to resolve diff mode according to context"""
    diff_mode = '[BT-diff]'
    if 'versus' in condition:
        diff_mode = '[OP-diff]_[BT-diff]'
    return diff_mode


def get_file_paths(subject, data_type, output, run_index=0,
                   onset='stim', mode='full',
                   sensor_mode='mag', hcp_path='.'):
    """This is the MNE-HCP file path synthesizer

    An easy conceptual mapper from questions to file paths

    Parameters
    ----------
    subject : str
        The subject, a 6 digit code.
    data_type : str
        The acquisition context of the data. The following ones are supported:
        'rest'
        'noise'
        'task_motor'
        'task_story_math'
        'task_working_memory'
    output : str
        The kind of output. The following ones are supported:
        'raw',
        'epochs'
        'evoked'
        'ica'
        'annotations'
        'trial_info'
        'freesurfer'
        'meg_anatomy'
    onset : {'stim', 'resp', 'sentence', 'block'}
        The event onset. Only considered for epochs and evoked outputs
        The mapping is generous, everything that is not a response is a
        stimulus, in the sense of internal or external events. sentence and
        block are specific to task_story_math.
    mode : {'full', 'minimal'}
        Return minimum files or all of them. Only relevant for freesurfer
        and anatomy
    sensor_mode : {'mag', 'planar'}
        The sensor projection. Defaults to 'mag'. Only relevant for
        evoked output.
    run_index : int
        The run index. For the first run, use 0, for the second, use 1.
        Also see HCP documentation for the number of runs for a given data
        type.
    hcp_path : str
        The HCP directory, defaults to op.curdir.

    Returns
    -------
    out : list of str
        The file names.
    """
    if data_type not in kind_map:
        raise ValueError('I never heard of `%s` -- are you sure this is a'
                         ' valid HCP type? I currenlty support:\n%s' % (
                             data_type, ' \n'.join(
                                 [k for k in kind_map if '_' in k])))
    context = ('rmeg' if 'rest' in data_type else 'tmeg')
    sensor_mode = evoked_map['modes'][sensor_mode]
    my_onset = _map_onset(onset, data_type, output)
    if data_type not in ('meg_anatomy', 'freesurfer'):
        my_runs = run_map[data_type]
        if run_index >= len(my_runs):
            raise ValueError('For `data_type=%s` we have %d runs. '
                             'You asked for run index %d.' % (
                                 data_type, len(my_runs), run_index))
        run_label = my_runs[run_index]
    else:
        run_label = None
    files = list()
    pipeline = pipeline_map.get(output, output)
    processing = 'preprocessed'
    if output == 'raw':
        processing = 'unprocessed'
    if processing == 'preprocessed':
        if (data_type in ('rest', 'noise_subject', 'noise_empty_room') and
                output in ('trial_info', 'evoked')):
            raise ValueError('%s not defined for %s' % (output, data_type))

        file_map = preprocessed[(data_type if data_type in (
                                 'meg_anatomy', 'freesurfer') else 'meg')]
        path = file_map['path'].format(
            subject=subject,
            pipeline=(context + 'preproc' if output == 'epochs'
                      else pipeline),
            kind=kind_map[data_type])

        if output == 'epochs':
            pattern_key = (output, context)
        else:
            pattern_key = output

        my_pattern = file_map['patterns'][pattern_key]

        if output in ('bads', 'ica'):
            files.extend(
                [op.join(path,
                         p.format(subject=subject, run=run_label,
                                  kind=kind_map[data_type]))
                 for p in my_pattern])

        elif output == 'epochs':
            if 'noise' in data_type:
                raise ValueError('You want preprocessed data of type "%s". '
                                 'But the HCP does not ship those. Sorry.' %
                                 data_type)
            my_pattern = my_pattern[0]
            formats = dict(
                subject=subject, run=run_label, kind=kind_map[data_type],
                context=context)
            if context != 'rest':
                formats.update(onset=my_onset)
            this_file = my_pattern.format(**formats)
            files.append(op.join(path, this_file))

        elif output == 'evoked':
            # XXX add evoked template checks
            for condition in evoked_map[data_type]:
                if my_onset not in condition:
                    continue
                this_file = my_pattern[0].format(
                    subject=subject, kind=kind_map[data_type],
                    condition=condition,
                    diff_modes=_map_diff_mode(condition, data_type),
                    sensor_mode=sensor_mode)
                files.append(op.join(path, this_file))
        elif output == 'trial_info':
            this_file = my_pattern[0].format(
                subject=subject, run=run_label, kind=kind_map[data_type])
            files.append(op.join(path, this_file))
        elif data_type == 'meg_anatomy':
            path = file_map['path'].format(subject=subject)
            files.extend([op.join(path, pa.format(subject=subject))
                          for pa in my_pattern])
        elif data_type == 'freesurfer':
            path = file_map['path'].format(subject=subject)
            for pa in my_pattern:
                if mode == 'minimal':
                    if pa not in file_modes['freesurfer']['minimal']:
                        continue
                files.append(
                    op.join(path, output, pa.format(subject=subject)))
        else:
            raise ValueError('I never heard of `data_type` "%s".' % output)

    elif processing == 'unprocessed':
        if output == 'trial_info':
            raise ValueError('`trial_info` only exists for preprocessed data')
        path = unprocessed['path'].format(
            subject=subject, kind=kind_map[data_type], pipeline=pipeline,
            run=run_label)
        files.extend([op.join(path, p) for p in unprocessed['patterns']])

    else:
        raise ValueError('`processing` %s should be "unprocessed"'
                         ' or "preprocessed"')
    return [op.join(hcp_path, pa) for pa in files]



# Author: Denis A. Engemann <denis.engemann@gmail.com>
# License: BSD (3-clause)

import os.path as op
import itertools as itt
import re

import numpy as np
import scipy.io as scio
from scipy import linalg

from mne import (EpochsArray, EvokedArray, pick_info,
                 rename_channels)
from mne.io.bti.bti import _get_bti_info, read_raw_bti
from mne.io import _loc_to_coil_trans
from mne.utils import logger

from .file_mapping import get_file_paths


_label_mapping = [
    ('E1', 'ECG+'),
    ('E3', 'VEOG+'),
    ('E5', 'HEOG+'),
    ('E63', 'EMG_LF'),
    ('E31', 'EMG_LH'),
    ('E2', 'ECG-'),
    ('E4', 'VEOG-'),
    ('E6', 'HEOG-'),
    ('E64', 'EMG_RF'),
    ('E32', 'EMG_RH')
]

_time_lock_mapping = dict(
    TRESP='resp',
    TEMG='resp',
    TIM='stim',
    TEV='stim',
    TFLA='stim',
    BSENT='stim',
    BU='stim'
)


def _parse_trans(string):
    """helper to parse transforms"""
    return np.array(string.replace('\n', '')
                          .strip('[] ')
                          .split(' '), dtype=float).reshape(4, 4)


def _parse_hcp_trans(fid, transforms, convert_to_meter):
    """"another helper"""
    contents = fid.read()
    for trans in contents.split(';'):
        if 'filename' in trans or trans == '\n':
            continue
        key, trans = trans.split(' = ')
        key = key.lstrip('\ntransform.')
        transforms[key] = _parse_trans(trans)
        if convert_to_meter:
            transforms[key][:3, 3] *= 1e-3  # mm to m
    if not transforms:
        raise RuntimeError('Could not parse the transforms.')


def _read_trans_hcp(fname, convert_to_meter):
    """Read + parse transforms
    subject_MEG_anatomy_transform.txt
    """
    transforms = dict()
    with open(fname) as fid:
        _parse_hcp_trans(fid, transforms, convert_to_meter)
    return transforms


def _read_landmarks_hcp(fname):
    """XXX parse landmarks currently not used"""
    out = dict()
    with open(fname) as fid:
        for line in fid:
            kind, data = line.split(' = ')
            kind = kind.split('.')[1]
            if kind == 'coordsys':
                out['coord_frame'] = data.split(';')[0].replace("'", "")
            else:
                data = data.split()
                for c in ('[', '];'):
                    if c in data:
                        data.remove(c)
                out[kind] = np.array(data, dtype=int) * 1e-3  # mm to m
    return out


def _get_head_model(head_model_fname):
    """helper to parse head model from matfile"""
    head_mat = scio.loadmat(head_model_fname, squeeze_me=False)
    pnts = head_mat['headmodel']['bnd'][0][0][0][0][0]
    faces = head_mat['headmodel']['bnd'][0][0][0][0][1]
    faces -= 1  # correct for Matlab's 1-based index
    return pnts, faces


def _read_bti_info(raw_fid, config):
    """helper to only access bti info from pdf file"""
    info, bti_info = _get_bti_info(
        pdf_fname=raw_fid, config_fname=config, head_shape_fname=None,
        rotation_x=0.0, translation=(0.0, 0.02, 0.11),
        ecg_ch='E31', eog_ch=('E63', 'E64'),
        convert=False,  # no conversion to neuromag coordinates
        rename_channels=False,  # keep native channel names
        sort_by_ch_name=False)  # do not change native order
    return info


def _read_raw_bti(raw_fid, config_fid, convert):
    """Convert and raw file from HCP input"""
    raw = read_raw_bti(  # no convrt + no rename for HCP compatibility
        raw_fid, config_fid, convert=convert, head_shape_fname=None,
        sort_by_ch_name=False, rename_channels=False, preload=False)

    return raw


def _check_raw_config_runs(raws, configs):
    """XXX this goes to tests later, currently not used """
    for raw, config in zip(raws, configs):
        assert op.split(raw)[0] == op.split(config)[0]
    run_str = set([configs[0].split('/')[-3]])
    for config in configs[1:]:
        assert set(configs[0].split('/')) - set(config.split('/')) == run_str


def _check_infos_trans(infos):
    """XXX this goes to tests later, currently not used"""
    chan_max_idx = np.argmax([c['nchan'] for c in infos])
    chan_template = infos[chan_max_idx]['ch_names']
    channels = [c['ch_names'] for c in infos]
    common_channels = set(chan_template).intersection(*channels)

    common_chs = [[c['chs'][c['ch_names'].index(ch)] for ch in common_channels]
                  for c in infos]
    dev_ctf_trans = [i['dev_ctf_t']['trans'] for i in infos]
    cns = [[c['ch_name'] for c in cc] for cc in common_chs]
    for cn1, cn2 in itt.combinations(cns, 2):
        assert cn1 == cn2
    # BTI stores data in head coords, as a consequence the coordinates
    # change across run, we apply the ctf->ctf_head transform here
    # to check that all transforms are correct.
    cts = [np.array([linalg.inv(_loc_to_coil_trans(c['loc'])).dot(t)
                    for c in cc])
           for t, cc in zip(dev_ctf_trans, common_chs)]
    for ct1, ct2 in itt.combinations(cts, 2):
        np.testing.assert_array_almost_equal(ct1, ct2, 12)


def read_raw(subject, data_type, run_index=0, hcp_path=op.curdir):
    """Read HCP raw data

    Parameters
    ----------
    subject : str, file_map
        The subject
    data_type : str
        The kind of data to read. The following options are supported:
        'rest'
        'task_motor'
        'task_story_math'
        'task_working_memory'
        'noise_empty_room'
        'noise_subject'
    run_index : int
        The run index. For the first run, use 0, for the second, use 1.
        Also see HCP documentation for the number of runs for a given data
        type.
    hcp_path : str
        The HCP directory, defaults to op.curdir.

    Returns
    -------
    raw : instance of mne.io.Raw
        The MNE raw object.
    """
    pdf, config = get_file_paths(
        subject=subject, data_type=data_type, output='raw',
        run_index=run_index, hcp_path=hcp_path)

    raw = _read_raw_bti(pdf, config, convert=False)
    return raw


def read_info(subject, data_type, run_index=0, hcp_path=op.curdir):
    """Read info from unprocessed data

    Parameters
    ----------
    subject : str, file_map
        The subject
    data_type : str
        The kind of data to read. The following options are supported:
        'rest'
        'task_motor'
        'task_story_math'
        'task_working_memory'
        'noise_empty_room'
        'noise_subject'
    run_index : int
        The run index. For the first run, use 0, for the second, use 1.
        Also see HCP documentation for the number of runs for a given data
        type.
    hcp_path : str
        The HCP directory, defaults to op.curdir.

    Returns
    -------
    info : instance of mne.io.meas_info.Info
        The MNE channel info object.

    .. note::
        HCP MEG does not deliver only 3 of the 5 task packages from MRI HCP.
    """
    raw, config = get_file_paths(
        subject=subject, data_type=data_type, output='raw',
        run_index=run_index, hcp_path=hcp_path)

    if not op.exists(raw):
        raw = None

    meg_info = _read_bti_info(raw, config)

    if raw is None:
        logger.info('Did not find Raw data. Guessing EMG, ECG and EOG '
                    'channels')
        rename_channels(meg_info, dict(_label_mapping))
    return meg_info


def read_epochs(subject, data_type, onset='TIM', run_index=0,
                    hcp_path=op.curdir, return_fixations_motor=False):
    """Read HCP processed data

    Parameters
    ----------
    subject : str, file_map
        The subject
    data_type : str
        The kind of data to read. The following options are supported:
        'rest'
        'task_motor'
        'task_story_math'
        'task_working_memory'
    onset : {'stim', 'resp', 'sentence', 'block'}
        The event onset. The mapping is generous, everything that is not a
        response is a stimulus, in the sense of internal or external events.
        `sentence` and `block` are specific to task_story_math.
    run_index : int
        The run index. For the first run, use 0, for the second, use 1.
        Also see HCP documentation for the number of runs for a given data
        type.
    hcp_path : str
        The HCP directory, defaults to op.curdir.
    return_fixations_motor : bool
        Weather to return fixations or regular trials. For motor data only.
        Defaults to False.
    Returns
    -------
    epochs : instance of mne.Epochs
        The MNE epochs. Note, these are pseudo-epochs in the case of
        onset == 'rest'.
    """
    info = read_info(subject=subject, data_type=data_type,
                         run_index=run_index, hcp_path=hcp_path)

    epochs_mat_fname = get_file_paths(
        subject=subject, data_type=data_type, output='epochs',
        run_index=run_index, hcp_path=hcp_path)[0]
    if data_type != 'task_motor':
        return_fixations_motor = None
    epochs = _read_epochs(epochs_mat_fname=epochs_mat_fname, info=info,
                          return_fixations_motor=return_fixations_motor)
    if data_type == 'task_motor':
        epochs.set_channel_types(
            {ch: 'emg' for ch in epochs.ch_names if 'EMG' in ch})
    return epochs


def _read_epochs(epochs_mat_fname, info, return_fixations_motor):
    """read the epochs from matfile"""
    data = scio.loadmat(epochs_mat_fname,
                        squeeze_me=True)['data']
    ch_names = [ch for ch in data['label'].tolist()]
    info['sfreq'] = data['fsample'].tolist()
    times = data['time'].tolist()[0]

    # deal with different event lengths
    if return_fixations_motor is not None:
        fixation_mask = data['trialinfo'].tolist()[:, 1] == 6
        if return_fixations_motor is False:
            fixation_mask = ~fixation_mask
        data = np.array(data['trial'].tolist()[fixation_mask].tolist())
    else:
        data = np.array(data['trial'].tolist().tolist())

    # warning: data are not chronologically ordered but
    # match the trial info
    events = np.zeros((len(data), 3), dtype=np.int)
    events[:, 0] = np.arange(len(data))
    events[:, 2] = 99  # all events
    # we leave it to the user to construct his events
    # as from the data['trialinfo'] arbitrary events can be constructed.
    # and it is task specific.
    this_info = _hcp_pick_info(info, ch_names)
    epochs = EpochsArray(data=data, info=this_info, events=events,
                         tmin=times.min())
    return epochs


def _hcp_pick_info(info, ch_names):
    """helper to subset info"""
    return pick_info(
        info, [info['ch_names'].index(ch) for ch in ch_names],
        copy=True)


def read_trial_info(subject, data_type, run_index=0, hcp_path=op.curdir):
    """Read information about trials

    Parameters
    ----------
    subject : str
        The HCP subject.
    data_type : str
        The kind of data to read. The following options are supported:
        'rest'
        'task_motor'
        'task_story_math'
        'task_working_memory'
    run_index : int
        The run index. For the first run, use 0, for the second, use 1.
        Also see HCP documentation for the number of runs for a given data
        type.
    hcp_path : str
        The HCP directory, defaults to op.curdir.
    Returns
    -------
    trial_info : dict
        The trial info including event labels, indices and times.
    """

    trial_info_mat_fname = get_file_paths(
        subject=subject, data_type=data_type,
        output='trial_info', run_index=run_index,
        hcp_path=hcp_path)[0]

    trl_info = _read_trial_info(trial_info_mat_fname=trial_info_mat_fname)
    return trl_info


def _read_trial_info(trial_info_mat_fname):
    """helper to read trial info"""
    # XXX FIXME index -1
    data = scio.loadmat(trial_info_mat_fname, squeeze_me=True)['trlInfo']
    out = dict()

    for idx, lock_name in enumerate(data['lockNames'].tolist()):
        key = _time_lock_mapping[lock_name]
        out[key] = dict(
            comments=data['trlColDescr'].tolist()[idx],
            codes=data['lockTrl'].tolist().tolist()[idx])

    return out


def _check_sorting_runs(candidates, id_char):
    """helper to ensure correct run-parsing and mapping"""
    run_idx = [f.find(id_char) for f in candidates]
    for config, idx in zip(candidates, run_idx):
        assert config[idx - 1].isdigit()
        assert not config[idx - 2].isdigit()
    runs = [int(f[idx - 1]) for f, idx in zip(candidates, run_idx)]
    return runs, candidates


def _parse_annotations_segments(segment_strings):
    """Read bad segments defintions from text file"""
    for char in '}]':  # multi line array definitions
        segment_strings = segment_strings.replace(
            char + ';', 'splitme'
        )
    split = segment_strings.split('splitme')
    out = dict()
    for entry in split:
        if len(entry) == 1 or entry == '\n':
            continue
        key, rest = entry.split(' = ')
        val = np.array(
            [k for k in [''.join([c for c in e if c.isdigit()])
             for e in rest.split()] if k.isdigit()], dtype=int)
        # reshape and map to Python index
        val = val.reshape(-1, 2) - 1
        out[key.split('.')[1]] = val
    return out


def read_annot(subject, data_type, run_index=0, hcp_path=op.curdir):
    """Read annotations for bad data and ICA.

    Parameters
    ----------
    subject : str, file_map
        The subject
    data_type : str
        The kind of data to read. The following options are supported:
        'rest'
        'task_motor'
        'task_story_math'
        'task_working_memory'
    run_index : int
        The run index. For the first run, use 0, for the second, use 1.
        Also see HCP documentation for the number of runs for a given data
        type.
    hcp_path : str
        The HCP directory, defaults to op.curdir.

    Returns
    -------
    out : dict
        The annotations.
    """
    bads_files = get_file_paths(
        subject=subject, data_type=data_type,
        output='bads', run_index=run_index,
        hcp_path=hcp_path)
    segments_fname = [k for k in bads_files if
                      k.endswith('baddata_badsegments.txt')][0]
    bads_fname = [k for k in bads_files if
                  k.endswith('baddata_badchannels.txt')][0]

    ica_files = get_file_paths(
        subject=subject, data_type=data_type,
        output='ica', run_index=run_index,
        hcp_path=hcp_path)
    ica_fname = [k for k in ica_files if k.endswith('icaclass_vs.txt')][0]

    out = dict()
    iter_fun = [
        ('channels', _parse_annotations_bad_channels, bads_fname),
        ('segments', _parse_annotations_segments, segments_fname),
        ('ica', _parse_annotations_ica, ica_fname)]

    for subtype, fun, fname in iter_fun:
        with open(fname, 'r') as fid:
            out[subtype] = fun(fid.read())

    return out


def read_ica(subject, data_type, run_index=0, hcp_path=op.curdir):
    """Read precomputed independent components from subject

    Parameters
    ----------
    subject : str, file_map
        The subject
    data_type : str
        The kind of data to read. The following options are supported:
        'rest'
        'task_motor'
        'task_story_math'
        'task_working_memory'
    run_index : int
        The run index. For the first run, use 0, for the second, use 1.
        Also see HCP documentation for the number of runs for a given data
        type.
    hcp_path : str
        The HCP directory, defaults to op.curdir.

    Returns
    -------
    mat : numpy structured array
        The ICA mat struct.
    """

    ica_files = get_file_paths(
        subject=subject, data_type=data_type,
        output='ica', run_index=run_index,
        hcp_path=hcp_path)
    ica_fname_mat = [k for k in ica_files if k.endswith('icaclass.mat')][0]

    mat = scio.loadmat(ica_fname_mat, squeeze_me=True)['comp_class']
    return mat


def _parse_annotations_bad_channels(bads_strings):
    """Read bad channel definitions from text file"""
    for char in '}]':
        bads_strings = bads_strings.replace(
            char + ';', 'splitme'
        )
    split = bads_strings.split('splitme')
    out = dict()
    for entry in split:
        if len(entry) == 1 or entry == '\n':
            continue
        key, rest = entry.split(' = ')
        val = [ch for ch in rest.split("'") if ch.isalnum()]
        out[key.split('.')[1]] = val
    return out


def _parse_annotations_ica(ica_strings):
    """Read bad channel definitions from text file"""
    # prepare splitting
    for char in '}]':  # multi line array definitions
        ica_strings = ica_strings.replace(
            char + ';', 'splitme'
        )
    # scalar variables
    match_inds = list()
    for match in re.finditer(';', ica_strings):
        ii = match.start()
        if ica_strings[ii - 1].isalnum():
            match_inds.append(ii)

    ica_strings = list(ica_strings)
    for ii in match_inds:
        ica_strings[ii] = 'splitme'
    ica_strings = ''.join(ica_strings)

    split = ica_strings.split('splitme')
    out = dict()
    for entry in split:
        if len(entry) == 1 or entry == '\n':
            continue
        key, rest = entry.split(' = ')
        if '[' in rest:
            sep = ' '
        else:
            sep = "'"
        val = [ch for ch in rest.split(sep) if ch.isalnum()]
        if all(v.isdigit() for v in val):
            val = [int(v) - 1 for v in val]  # map to Python index
        out[key.split('.')[1]] = val
    return out


def read_evokeds(subject, data_type, onset='stim', sensor_mode='mag',
                     hcp_path=op.curdir, kind='average'):
    """Read HCP processed data

    Parameters
    ----------
    subject : str, file_map
        The subject
    data_type : str
        The kind of data to read. The following options are supported:
        'rest'
        'task_motor'
        'task_story_math'
        'task_working_memory'
    onset : {'stim', 'resp'}
        The event onset. The mapping is generous, everything that is not a
        response is a stimulus, in the sense of internal or external events.
    sensor_mode : {'mag', 'planar'}
        The sensor projection. Defaults to 'mag'. Only relevant for
        evoked output.
    hcp_path : str
        The HCP directory, defaults to op.curdir.
    kind : {'average', 'standard_error'}
        The averaging mode. Defaults to 'average'.
    Returns
    -------
    epochs : instance of mne.Epochs
        The MNE epochs. Note, these are pseudo-epochs in the case of
        onset == 'rest'.
    """
    info = read_info(subject=subject, data_type=data_type,
                         hcp_path=hcp_path, run_index=0)

    evoked_files = list()
    for fname in get_file_paths(
            subject=subject, data_type=data_type, onset=onset,
            output='evoked', sensor_mode=sensor_mode, hcp_path=hcp_path):
        evoked_files.extend(_read_evoked(fname, sensor_mode, info, kind))
    return evoked_files


def _read_evoked(fname, sensor_mode, info, kind):
    """helper to read evokeds"""
    data = scio.loadmat(fname, squeeze_me=True)['data']
    ch_names = [ch for ch in data['label'].tolist()]

    times = data['time'].tolist()
    sfreq = 1. / np.diff(times)[0]

    info = _hcp_pick_info(info, ch_names)
    info['sfreq'] = sfreq

    out = list()
    comment = ('_'.join(fname.split('/')[-1].split('_')[2:])
                  .replace('.mat', '')
                  .replace('_eravg_', '_')
                  .replace('[', '')
                  .replace(']', ''))
    nave = np.unique(data['dof'].tolist())
    assert len(nave) == 1
    nave = nave[0]
    for key, this_kind in (('var', 'standard_error'), ('avg', 'average')):
        if this_kind != kind:
            continue
        evoked = EvokedArray(
            data=data[key].tolist(), info=info, tmin=min(times),
            kind=this_kind, comment=comment, nave=nave)
        out.append(evoked)
    return out



# Author: Denis A. Engemann <denis.engemann@gmail.com>
# License: BSD (3-clause)

import os
import os.path as op
from ..io.file_mapping import get_s3_keys_anatomy, get_s3_keys_meg


hcp_prefix = 's3://hcp-openaccess/HCP_900'
subject = '105923'


hcp_data_types = [
    'rest',
    'task_working_memory',
    'task_story_math',
    'task_motor',
    'noise_empty_room'
]


hcp_outputs = [
    'raw',
    'epochs',
    'ica',
    'evoked',
    'trial_info',
    'bads'
]

hcp_cheap = os.getenv('MNE_HCP_CHEAP', False)

hcp_onsets = ['stim']

# allow for downloading fewer data
run_inds = [0, 1, 2]
max_runs = int(os.getenv('MNE_HCP_N_RUNS', 3))
s3_keys = list()

s3_keys += get_s3_keys_meg(
    subject,
    data_types=hcp_data_types,
    onsets=hcp_onsets,
    hcp_path_bucket=hcp_prefix,
    outputs=[dd for dd in hcp_outputs if dd in ('raw',)],
    run_inds=run_inds[:max_runs])

if hcp_cheap:
    s3_keys = [kk for kk in s3_keys if
               'Rest' in kk or ('Rest' not in kk and 'config' in kk)]

s3_keys += get_s3_keys_meg(
    subject,
    data_types=hcp_data_types,
    onsets=hcp_onsets,
    hcp_path_bucket=hcp_prefix,
    outputs=[dd for dd in hcp_outputs if dd in ('epochs')],
    run_inds=run_inds[:max_runs])

s3_keys += get_s3_keys_meg(
    subject,
    data_types=hcp_data_types,
    onsets=hcp_onsets,
    hcp_path_bucket=hcp_prefix,
    outputs=[dd for dd in hcp_outputs if dd not in ('raw', 'epochs')],
    run_inds=run_inds)

s3_keys += get_s3_keys_anatomy(
    subject, hcp_path_bucket=hcp_prefix, mode='minimal')

##############################################################################
# variable used in different tests

hcp_path = op.expanduser('~/mne-hcp-data/HCP')

bti_chans = {'A' + str(i) for i in range(1, 249, 1)}

test_subject = '105923'
task_types = ['task_story_math', 'task_working_memory', 'task_motor']
noise_types = ['noise_empty_room']
sfreq_preproc = 508.63
sfreq_raw = 2034.5101
lowpass_preproc = 150
highpass_preproc = 1.3

epochs_bounds = {
    'task_motor': (-1.2, 1.2),
    'task_working_memory': (-1.5, 2.5),
    'task_story_math': (-1.5, 4),
    'rest': (0, 2)
}


def nottest(f):
    """Decorator to mark a function as not a test"""
    f.__test__ = False
    return f


@nottest
def expensive_test(f):
    """Decorator for expensive testing"""
    f.expensive_test = True
    return f



# Author: Denis A. Engemann <denis.engemann@gmail.com>
# License: BSD (3-clause)

import numpy as np
import mne
from mne.io import set_bipolar_reference
from mne.io.bti.bti import (
    _convert_coil_trans, _coil_trans_to_loc, _get_bti_dev_t,
    _loc_to_coil_trans)
from mne.transforms import Transform
from mne.utils import logger

from .io import read_info
from .io.read import _hcp_pick_info


def set_eog_ecg_channels(raw):
    """Set the HCP ECG and EOG channels

    .. note::
       Operates in place.

    Parameters
    ----------
    raw : instance of Raw
        the hcp raw data.
    """
    for kind in ['ECG', 'VEOG', 'HEOG']:
        set_bipolar_reference(
            raw, anode=kind + '-', cathode=kind + '+', ch_name=kind,
            copy=False)
    raw.set_channel_types({'ECG': 'ecg', 'VEOG': 'eog', 'HEOG': 'eog'})


def apply_ica_hcp(raw, ica_mat, exclude):
    """Apply the HCP ICA.

    .. note::
       Operates in place.

    Parameters
    ----------
    raw : instance of Raw
        the hcp raw data.
    ica_mat : numpy structured array
        The hcp ICA solution
    exclude : array-like
        the components to be excluded.
    """
    ch_names = ica_mat['topolabel'].tolist().tolist()
    picks = mne.pick_channels(raw.info['ch_names'], include=ch_names)
    assert ch_names == [raw.ch_names[p] for p in picks]

    unmixing_matrix = np.array(ica_mat['unmixing'].tolist())

    n_components, n_channels = unmixing_matrix.shape
    mixing = np.array(ica_mat['topo'].tolist())

    proj_mat = (np.eye(n_channels) - np.dot(
        mixing[:, exclude], unmixing_matrix[exclude]))
    raw._data *= 1e15
    raw._data[picks] = np.dot(proj_mat, raw._data[picks])
    raw._data /= 1e15


def apply_ref_correction(raw, decim_fit=100):
    """Regress out MEG ref channels

    Computes linear models from MEG reference channels
    on each sensors, predicts the MEG data and subtracts
    and computes the residual by subtracting the predictions.

    .. note::
       Operates in place.

    .. note::
       Can be memory demanding. To alleviate this problem the model can be fit
       on decimated data. This is legitimate because the linear model does
       not have any representation of time, only the distributions
       matter.

    Parameters
    ----------
    raw : instance of Raw
        The BTi/4D raw data.
    decim_fit : int
        The decimation factor used for fitting the model.
        Defaults to 100.
    """
    from sklearn.linear_model import LinearRegression

    meg_picks = mne.pick_types(raw.info, ref_meg=False, meg=True)
    ref_picks = mne.pick_types(raw.info, ref_meg=True, meg=False)
    if len(ref_picks) == 0:
        raise ValueError('Could not find meg ref channels.')

    estimator = LinearRegression(normalize=True)  # ref MAG + GRAD
    Y_pred = estimator.fit(
        raw[ref_picks][0][:, ::decim_fit].T,
        raw[meg_picks][0][:, ::decim_fit].T).predict(
        raw[ref_picks][0].T)
    raw._data[meg_picks] -= Y_pred.T


def map_ch_coords_to_mne(inst):
    """Transform sensors to MNE coordinates

    .. note::
        operates in place

    .. warning::
        For several reasons we do not use the MNE coordinates for the inverse
        modeling. This however won't always play nicely with visualization.

    Parameters
    ----------
    inst :  MNE data containers
        Raw, Epochs, Evoked.
    """
    bti_dev_t = Transform('ctf_meg', 'meg', _get_bti_dev_t())
    dev_ctf_t = inst.info['dev_ctf_t']
    for ch in inst.info['chs']:
        loc = ch['loc'][:]
        if loc is not None:
            logger.debug('converting %s' % ch['ch_name'])
            t = _loc_to_coil_trans(loc)
            t = _convert_coil_trans(t, dev_ctf_t, bti_dev_t)
            loc = _coil_trans_to_loc(t)
            ch['loc'] = loc


def interpolate_missing(inst, subject, data_type, hcp_path,
                        run_index=0, mode='fast'):
    """Interpolate all MEG channels that are missing

    .. warning::
       This function may require some memory.

    Parameters
    ----------
    inst :  MNE data containers
        Raw, Epochs, Evoked.
    subject : str, file_map
        The subject
    data_type : str
        The kind of data to read. The following options are supported:
        'rest'
        'task_motor'
        'task_story_math'
        'task_working_memory'
        'noise_empty_room'
        'noise_subject'
    run_index : int
        The run index. For the first run, use 0, for the second, use 1.
        Also see HCP documentation for the number of runs for a given data
        type.
    hcp_path : str
        The HCP directory, defaults to op.curdir.
    mode : str
        Either `'accurate'` or `'fast'`, determines the quality of the
        Legendre polynomial expansion used for interpolation of MEG
        channels.

    Returns
    -------
    out :   MNE data containers
        Raw, Epochs, Evoked but with missing channels interpolated.
    """
    try:
        info = read_info(
            subject=subject, data_type=data_type, hcp_path=hcp_path,
            run_index=run_index if run_index is None else run_index)
    except (ValueError, IOError):
        raise ValueError(
            'could not find config to complete info.'
            'reading only channel positions without '
            'transforms.')
    # figure out which channels are missing
    bti_channel_names_ = ['A%i' % ii for ii in range(1, 249, 1)]
    bti_channel_names = [ch for ch in info['ch_names']
                         if ch in bti_channel_names_]
    fake_channels_to_add = sorted(
        list(set(bti_channel_names) - set(inst.ch_names)))
    n_channels = 248
    info['sfreq'] = inst.info['sfreq']
    # compute shape of data to be added
    is_raw = isinstance(inst, (mne.io.Raw,
                        mne.io.RawArray,
                        mne.io.bti.bti.RawBTi))
    is_epochs = isinstance(inst, (mne.Epochs, mne.EpochsArray))
    is_evoked = isinstance(inst, (mne.Evoked, mne.EvokedArray))
    if is_raw:
        shape = (n_channels,
                 (inst.last_samp - inst.first_samp) + 1)
        data = inst._data
    elif is_epochs:
        shape = (n_channels, len(inst.events), len(inst.times))
        data = np.transpose(inst.get_data(), (1, 0, 2))
    elif is_evoked:
        shape = (n_channels, len(inst.times))
        data = inst.data
    else:
        raise ValueError('instance must be Raw, Epochs '
                         'or Evoked')

    # create new data
    existing_channels_index = [
        bti_channel_names.index(ch) for ch in inst.ch_names]
    new_channels_index = [
        bti_channel_names.index(ch) for ch in fake_channels_to_add]

    out_data = np.empty(shape, dtype=np.float64)
    out_data[existing_channels_index] = data
    out_data[new_channels_index] = 0
    info = _hcp_pick_info(info, bti_channel_names)

    if is_raw:
        out = mne.io.RawArray(out_data, info)
        if inst.annotations is not None:
            out.annotations = inst.annotations
    elif is_epochs:
        out = mne.EpochsArray(data=np.transpose(out_data, (1, 0, 2)),
                              info=info, events=inst.events,
                              tmin=inst.times.min(), event_id=inst.event_id)
    elif is_evoked:
        out = mne.EvokedArray(
            data=out_data, info=info, tmin=inst.times.min(),
            comment=inst.comment, nave=inst.nave, kind=inst.kind)
    else:
        raise ValueError('instance must be Raw, Epochs '
                         'or Evoked')

    # set "bad" channels and interpolate.
    out.info['bads'] = fake_channels_to_add
    out.interpolate_bads(mode=mode)
    return out



import os.path as op

import numpy as np

import mne
from mne.io.pick import _pick_data_channels, pick_info
from mne import read_trans, read_surface
from mne.transforms import apply_trans
from mne.viz.topomap import _find_topomap_coords

from .io import read_info


def make_hcp_bti_layout(info):
    """Get Layout of HCP Magnes3600WH data

    Parameters
    ----------
    info : mne.io.meas_info.Info
        The measurement info.

    Returns
    -------
    lout : mne.channels.Layout
        The layout that can be used for plotting.
    """
    picks = list(range(248))
    pos = _find_topomap_coords(info, picks=picks)
    return mne.channels.layout.Layout(
        box=(-42.19, 43.52, -41.7, 28.71), pos=pos,
        names=[info['ch_names'][idx] for idx in picks], ids=picks,
        kind='magnesWH3600_hcp')


def plot_coregistration(subject, subjects_dir, recordings_path,
                        hcp_path=op.curdir,
                        info_from=(('data_type', 'rest'), ('run_index', 0)),
                        view_init=(('azim', 0), ('elev', 0))):
    """A diagnostic plot to show the HCP coregistration

    Parameters
    ----------

    subject : str
        The subject
    subjects_dir : str
        The path corresponding to MNE/freesurfer SUBJECTS_DIR (to be created)
    hcp_path : str
        The path where the HCP files can be found. defaults to op.curdir.
    info_from : tuple of tuples | dict
        The reader info concerning the data from which sensor positions
        should be read.
        Must not be empty room as sensor positions are in head
        coordinates for 4D systems, hence not available in that case.
        Note that differences between the sensor positions across runs
        are smaller than 12 digits, hence negligible.
    view_init : tuple of tuples | dict
        The initival view, defaults to azimuth and elevation of 0,
        a simple lateral view

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object.
    """
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D  #  noqa

    if isinstance(info_from, tuple):
        info_from = dict(info_from)
    if isinstance(view_init, tuple):
        view_init = dict(view_init)

    head_mri_t = read_trans(
        op.join(recordings_path, subject,
                '{}-head_mri-trans.fif'.format(subject)))

    info = read_info(subject=subject, hcp_path=hcp_path, **info_from)

    info = pick_info(info, _pick_data_channels(info, with_ref_meg=False))
    sens_pnts = np.array([c['loc'][:3] for c in info['chs']])
    sens_pnts = apply_trans(head_mri_t, sens_pnts)
    sens_pnts *= 1e3  # put in mm scale

    pnts, tris = read_surface(
        op.join(subjects_dir, subject, 'bem', 'inner_skull.surf'))

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(*sens_pnts.T, color='purple', marker='o')
    ax.scatter(*pnts.T, color='green', alpha=0.3)
    ax.view_init(**view_init)
    fig.tight_layout()
    return fig



#! /usr/bin/env python
#

# Copyright (C) 2015-2016 Denis Engemann
# <denis.engemann@gmail.com>

import os
from os import path as op

import setuptools  # noqa; we are using a setuptools namespace
from numpy.distutils.core import setup

# get the version (don't import mne here, so dependencies are not needed)
version = None
with open(os.path.join('hcp', '__init__.py'), 'r') as fid:
    for line in (line.strip() for line in fid):
        if line.startswith('__version__'):
            version = line.split('=')[1].strip().strip('\'')
            break
if version is None:
    raise RuntimeError('Could not determine version')


descr = """MNE HCP project for accessing the human connectome MEG data in Python.""" # noqa

DISTNAME = 'mne-hcp'
DESCRIPTION = descr
MAINTAINER = 'Denis A. Engemann'
MAINTAINER_EMAIL = 'denis.engemann@gmail.com'
URL = 'http://github.com/mne-tools/mne-hcp'
LICENSE = 'BSD (3-clause)'
DOWNLOAD_URL = 'http://github.com/mne-tools/mne-hcp'
VERSION = version


if __name__ == "__main__":
    if os.path.exists('MANIFEST'):
        os.remove('MANIFEST')

    setup(name=DISTNAME,
          maintainer=MAINTAINER,
          include_package_data=True,
          maintainer_email=MAINTAINER_EMAIL,
          description=DESCRIPTION,
          license=LICENSE,
          url=URL,
          version=VERSION,
          download_url=DOWNLOAD_URL,
          long_description=open('README.rst').read(),
          zip_safe=False,  # the package can run out of an .egg file
          classifiers=['Intended Audience :: Science/Research',
                       'Intended Audience :: Developers',
                       'License :: OSI Approved',
                       'Programming Language :: Python',
                       'Topic :: Software Development',
                       'Topic :: Scientific/Engineering',
                       'Operating System :: Microsoft :: Windows',
                       'Operating System :: POSIX',
                       'Operating System :: Unix',
                       'Operating System :: MacOS'],
          platforms='any',
          packages=['hcp',
                    'hcp.io',
                    'hcp.io.tests',
                    'hcp.io.file_mapping',
                    'hcp.io.file_mapping.tests',
                    'hcp.tests'],
          package_data={'hcp': [
              op.join('io', 'file_mapping', 'data', '*txt')]})



import shutil
import os
import os.path as op

import numpy as np

from numpy.testing import assert_array_equal
from nose.tools import assert_equal, assert_true, assert_raises


import mne
import hcp
from mne.utils import _TempDir
from hcp.tests import config as tconf
from hcp.tests.config import expensive_test
from hcp.io.read import _hcp_pick_info

hcp_params = dict(hcp_path=tconf.hcp_path,
                  subject=tconf.test_subject)


def test_read_annot():
    """testing annotations"""
    for run_index in tconf.run_inds:
        annots = hcp.read_annot(data_type='rest', run_index=run_index,
                                **hcp_params)
        # channels
        assert_equal(list(sorted(annots['channels'])),
                     ['all', 'ica', 'manual',  'neigh_corr',
                      'neigh_stdratio'])
        for channels in annots['channels'].values():
            for chan in channels:
                assert_true(chan in tconf.bti_chans)

        # segments
        assert_equal(list(sorted(annots['ica'])),
                     ['bad', 'brain_ic', 'brain_ic_number',
                      'brain_ic_vs', 'brain_ic_vs_number',
                      'ecg_eog_ic', 'flag', 'good',
                      'physio', 'total_ic_number'])
        for components in annots['ica'].values():
            if len(components) > 0:
                assert_true(min(components) >= 0)
                assert_true(max(components) <= 248)


def _basic_raw_checks(raw):
    """helper for testing raw files """
    picks = mne.pick_types(raw.info, meg=True, ref_meg=False)
    assert_equal(len(picks), 248)
    ch_names = [raw.ch_names[pp] for pp in picks]
    assert_true(all(ch.startswith('A') for ch in ch_names))
    ch_sorted = list(sorted(ch_names))
    assert_true(ch_sorted != ch_names)
    assert_equal(np.round(raw.info['sfreq'], 4),
                 tconf.sfreq_raw)


def test_read_raw_rest():
    """Test reading raw for resting state"""
    for run_index in tconf.run_inds[:tconf.max_runs]:
        raw = hcp.read_raw(data_type='rest', run_index=run_index,
                           **hcp_params)
        _basic_raw_checks(raw=raw)


@expensive_test
def test_read_raw_task():
    """Test reading raw for tasks"""
    for run_index in tconf.run_inds[:tconf.max_runs]:
        for data_type in tconf.task_types:
            if run_index == 2:
                assert_raises(
                    ValueError, hcp.read_raw,
                    data_type=data_type, run_index=run_index, **hcp_params)
                continue
            raw = hcp.read_raw(
                data_type=data_type, run_index=run_index, **hcp_params)
            _basic_raw_checks(raw=raw)


@expensive_test
def test_read_raw_noise():
    """Test reading raw for empty room noise"""
    for run_index in tconf.run_inds[:tconf.max_runs][:2]:
        for data_type in tconf.noise_types:
            if run_index == 1:
                assert_raises(
                    ValueError, hcp.read_raw,
                    data_type=data_type, run_index=run_index, **hcp_params)
                continue
            raw = hcp.read_raw(
                data_type=data_type, run_index=run_index, **hcp_params)
            _basic_raw_checks(raw=raw)


def _epochs_basic_checks(epochs, annots, data_type):
    n_good = 248 - len(annots['channels']['all'])
    if data_type == 'task_motor':
        n_good += 4
    assert_equal(len(epochs.ch_names), n_good)
    assert_equal(
        np.round(epochs.info['sfreq'], 2),
        tconf.sfreq_preproc)
    assert_array_equal(
        np.unique(epochs.events[:, 2]),
        np.array([99], dtype=np.int))
    assert_true(
        _check_bounds(epochs.times,
                      tconf.epochs_bounds[data_type])
    )

    # XXX these seem not to be reliably set. checkout later.
    # assert_equal(
    #     epochs.info['lowpass'],
    #     lowpass_preproc)
    # assert_equal(
    #     epochs.info['highpass'],
    #     highpass_preproc)


def test_read_epochs_rest():
    """Test reading epochs for resting state"""
    for run_index in tconf.run_inds[:tconf.max_runs][:2]:
        annots = hcp.read_annot(
            data_type='rest', run_index=run_index, **hcp_params)

        epochs = hcp.read_epochs(
            data_type='rest', run_index=run_index, **hcp_params)

        _epochs_basic_checks(epochs, annots, data_type='rest')


def test_read_epochs_task():
    """Test reading epochs for task"""
    for run_index in tconf.run_inds[:tconf.max_runs][:2]:
        for data_type in tconf.task_types:
            annots = hcp.read_annot(
                data_type=data_type, run_index=run_index, **hcp_params)

            epochs = hcp.read_epochs(
                data_type=data_type, run_index=run_index, **hcp_params)

            _epochs_basic_checks(epochs, annots, data_type)


def _check_bounds(array, bounds):
    """helper for bounds checking"""
    is_in = True
    if not np.allclose(np.min(array), min(bounds), atol=0.01):
        is_in = False
    elif not np.allclose(np.max(array), max(bounds), atol=0.01):
        is_in = False
    return is_in


def test_read_evoked():
    """test reading evokeds """
    for data_type in tconf.task_types:
        all_annots = list()
        for run_index in tconf.run_inds[:2]:
            annots = hcp.read_annot(
                data_type=data_type, run_index=run_index, **hcp_params)
            all_annots.append(annots)

        evokeds = hcp.read_evokeds(data_type=data_type,
                                   kind='average', **hcp_params)

        n_average = sum(ee.kind == 'average' for ee in evokeds)
        assert_equal(n_average, len(evokeds))

        n_chans = 248
        if data_type == 'task_motor':
            n_chans += 4
        n_chans -= len(set(sum(
            [an['channels']['all'] for an in all_annots], [])))
        assert_equal(n_chans, len(evokeds[0].ch_names))
        assert_true(
            _check_bounds(evokeds[0].times,
                          tconf.epochs_bounds[data_type])
        )


def test_read_info():
    """test reading info"""
    tempdir = _TempDir()
    for data_type in tconf.task_types:
        for run_index in tconf.run_inds[:tconf.max_runs][:2]:
            # with pdf file
            info = hcp.read_info(
                data_type=data_type, run_index=run_index, **hcp_params)
            assert_equal(
                {k for k in info['ch_names'] if k.startswith('A')},
                tconf.bti_chans
            )
            # without pdf file
            # in this case the hcp code guesses certain channel labels
            cp_paths = hcp.io.file_mapping.get_file_paths(
                subject=tconf.test_subject, data_type=data_type,
                run_index=run_index, output='raw', hcp_path='',
            )
            for pp in cp_paths:
                if 'c,' in pp:  # don't copy pdf
                    continue
                os.makedirs(op.join(tempdir, op.dirname(pp)))
                shutil.copy(op.join(tconf.hcp_path, pp), op.join(tempdir, pp))

            info2 = hcp.read_info(
                subject=tconf.test_subject, data_type=data_type,
                hcp_path=tempdir,
                run_index=run_index)
            if os.getenv('MNE_HCP_CHEAP', False):
                assert_true(len(info['chs']) == len(info2['chs']))
            else:
                assert_true(len(info['chs']) != len(info2['chs']))
            common_chs = [ch for ch in info2['ch_names'] if
                          ch in info['ch_names']]
            assert_equal(len(common_chs), len(info['chs']))
            info2 = _hcp_pick_info(info2, common_chs)
            assert_equal(info['ch_names'], info2['ch_names'])
            for ch1, ch2 in zip(info['chs'], info2['chs']):
                assert_array_equal(ch1['loc'], ch2['loc'])


def test_read_trial_info():
    """Test trial info basics"""
    for data_type in tconf.task_types:
        for run_index in tconf.run_inds[:tconf.max_runs][:2]:
            trial_info = hcp.read_trial_info(
                data_type=data_type, run_index=run_index, **hcp_params)
            assert_true('stim' in trial_info)
            assert_true('resp' in trial_info)
            assert_equal(2, len(trial_info))
            for key, val in trial_info.items():
                assert_array_equal(np.ndim(val['comments']), 1)
                assert_array_equal(np.ndim(val['codes']), 2)
